import json
from abc import abstractmethod, ABCMeta
from logging import Logger

import requests
import urllib3
from requests import Session

from sidecar.const import Const
from sidecar.kub_api_pod_reader import KubApiPodReader
from sidecar.kub_token_provider import KubTokenProvider


class IKubApiService(metaclass=ABCMeta):
    def __init__(self) -> None:
        pass

    @abstractmethod
    def get_all_pods_list(self, include_infra: bool = True, include_ended=True, include_terminating: bool = True) -> []:
        raise NotImplementedError

    @abstractmethod
    def try_get_pod_json_by_ip(self, ip_address: str) -> {}:
        raise NotImplementedError

    @abstractmethod
    def try_get_pod_json_by_container_id(self, container_id: str) -> {}:
        raise NotImplementedError

    @abstractmethod
    def try_get_pod_json_by_pod_name(self, pod_name: str) -> {}:
        raise NotImplementedError

    @abstractmethod
    def update_namespace(self, annotations: {}):
        raise NotImplementedError

    @abstractmethod
    def update_pod(self, pod_name: str, annotations: {}):
        raise NotImplementedError

    @abstractmethod
    def update_service(self, name: str, annotations: {}):
        raise NotImplementedError

    @abstractmethod
    def get_all_services(self) -> {}:
        raise NotImplementedError


class KubApiService(IKubApiService):
    def __init__(self, hostname: str, namespace: str, kub_token_provider: KubTokenProvider, logger: Logger):
        super().__init__()
        self.logger = logger
        self.namespace = namespace
        self.hostname = hostname
        self._disable_secure_warnings()
        self.token = kub_token_provider.get_token()

    def get_all_pods_list(self, include_infra: bool = True, include_ended=True, include_terminating: bool = True) -> []:
        s = self._get_session()
        s.headers['Content-Type'] = 'application/strategic-merge-patch+json'
        url = '{hostname}/api/v1/namespaces/{namespace}/pods'. \
            format(hostname=self.hostname,
                   namespace=self.namespace)
        query_params = {}
        if not include_infra:
            query_params['labelSelector'] = '{app_selector_label}!={sidecar_selector_value}'.format(
                app_selector_label=Const.K8S_SIDECAR_APP_SELECTOR,
                sidecar_selector_value=Const.K8S_SIDECAR_SERVICE)
        res = s.get(url=url, verify=False, params=query_params)
        try:
            res.raise_for_status()
        except Exception as exc:
            self.logger.exception("Got exception while getting all pods list, message: {message}".format(message=str(exc)))
            raise exc

        pods_json = res.json()
        pods = pods_json['items']
        live_app_pods = self._filter_pods(pods=pods, include_ended=include_ended,
                                          include_terminating=include_terminating)
        return live_app_pods

    def try_get_pod_json_by_ip(self, ip_address: str) -> {}:
        # when we're looking for a specific pod, we want to find it no matter in what state it is
        pods = self.get_all_pods_list()
        pod_json = next(iter([pod for pod in pods if KubApiPodReader.safely_get_pod_ip(pod) == ip_address]), None)
        return pod_json

    def try_get_pod_json_by_container_id(self, container_id: str) -> {}:
        # when we're looking for a specific pod, we want to find it no matter in what state it is
        pods = self.get_all_pods_list()
        pod_json = next(iter([pod for pod in pods
                              if container_id in KubApiPodReader.safely_get_container_ids_in_pod(pod=pod)]),
                        None)
        return pod_json

    def try_get_pod_json_by_pod_name(self, pod_name: str) -> {}:
        # when we're looking for a specific pod, we want to find it no matter in what state it is
        s = self._get_session()
        s.headers['Content-Type'] = 'application/strategic-merge-patch+json'
        url = '{hostname}/api/v1/namespaces/{namespace}/pods/{pod_name}'. \
            format(hostname=self.hostname,
                   namespace=self.namespace,
                   pod_name=pod_name)
        res = s.get(url=url, verify=False)
        if res.status_code == 404:
            self.logger.error("Got 404 while trying to try_get_pod_json_by_pod_name, pod name: {pod_name}"
                              .format(pod_name=pod_name))
            return None
        try:
            res.raise_for_status()
        except Exception as exc:
            self.logger.exception("Got exception while getting json by pd name, pod name: {pod_name} , message: {message}".format(pod_name=pod_name, message=str(exc)))
            raise exc
        pod_json = res.json()
        return pod_json

    def get_all_services(self) -> []:
        s = self._get_session()
        url = '{}/api/v1/namespaces/{}/services/'.format(self.hostname, self.namespace)
        res = s.get(url=url, verify=False)
        if res.status_code == 404:
            self.logger.error("Could not find services in namespace {}".format(self.namespace))
            return None
        try:
            res.raise_for_status()
            return res.json()['items']
        except Exception as exc:
            self.logger.exception(exc)
            raise exc

    def _update(self, url: str, annotations: {}):
        s = self._get_session()
        s.headers['Content-Type'] = 'application/strategic-merge-patch+json'
        data_json = {
            'metadata': {
                'annotations': annotations
            }
        }
        metadata_json = json.dumps(data_json)
        res = s.patch(url=url, data=metadata_json, verify=False)
        res.raise_for_status()

    def update_service(self, name: str, annotations: {}):
        url = '{hostname}/api/v1/namespaces/{namespace}/services/{name}' \
            .format(hostname=self.hostname,
                    namespace=self.namespace,
                    name=name)
        try:
            self._update(url=url, annotations=annotations)
        except Exception as exc:
            self.logger.exception("Got exception while updating service, message: {message}".format(message=str(exc)))
            raise exc

    def update_namespace(self, annotations: {}):
        url = '{hostname}/api/v1/namespaces/{namespace}'.format(hostname=self.hostname,
                                                                namespace=self.namespace)
        try:
            self._update(url=url, annotations=annotations)
        except Exception as exc:
            self.logger.exception("Got exception while updating namespace, message: {message}".format(message=str(exc)))
            raise exc

    def update_pod(self, pod_name: str, annotations: {}):
        url = '{hostname}/api/v1/namespaces/{namespace}/pods/{pod_name}' \
            .format(hostname=self.hostname,
                    namespace=self.namespace,
                    pod_name=pod_name)
        try:
            self._update(url=url, annotations=annotations)
        except Exception as exc:
            self.logger.exception("Got exception while updating pod, pod name: {pod_name}, message: {message}".format(pod_name=pod_name, message=str(exc)))
            raise exc

    def _get_session(self) -> Session:
        s = requests.Session()
        s.headers['Authorization'] = 'Bearer ' + self.token
        return s

    @staticmethod
    def _filter_pods(pods: [], include_ended: bool, include_terminating: bool) -> []:
        # filtering out ended pods relies on the assumption that pods' restart policy is Always,
        # meaning that new pods were created instead of the ended ones and the old ones are no longer relevant
        # the terminating pods are filtered out because under this restart policy sometimes the new pod can be created
        # while the old one is still terminating and we don't want to get both of them, but only the "live" one
        return [pod for pod in pods
                if (include_ended or not KubApiPodReader.is_pod_ended(pod)) and
                (include_terminating or not KubApiPodReader.is_pod_terminating(pod))
                ]

    @staticmethod
    def _disable_secure_warnings():
        urllib3.disable_warnings(urllib3.exceptions.InsecureRequestWarning)



from typing import List, Dict

from sidecar.app_request_info import AppRequestInfo
from sidecar.apps_configuration_end_tracker import AppsConfigurationEndTracker, AppConfigurationEndStatus
from sidecar.const import AppNetworkStatus
from sidecar.health_check.app_health_check_state import AppHealthCheckState


class AppConfigurationStartStatus:
    START = "start"
    WAIT = "wait"


class AppConfigurationStartPolicy:
    def __init__(self,
                 app_health_check_state: AppHealthCheckState,
                 apps_config_end_tracker: AppsConfigurationEndTracker,
                 app_requests: List[AppRequestInfo]):
        self._app_health_check_state = app_health_check_state
        self._apps_config_end_tracker = apps_config_end_tracker
        self._app_dependencies = self._build_app_dependencies(app_requests)

    def get_app_configuration_start_status(self, app_name: str) -> str:
        app_dependency_names = self._app_dependencies[app_name]
        if not app_dependency_names:
            return AppConfigurationStartStatus.START

        app_dependency_statuses = self._apps_config_end_tracker.get_app_configuration_statuses(*app_dependency_names)
        all_instances_completed_with_success = all(
            config_end_status.is_ended_with_status(AppConfigurationEndStatus.COMPLETED) for config_end_status in
            app_dependency_statuses.values())

        all_apps_completed_private_network_health_check = all(
            AppNetworkStatus.passed_internal_network_test(status)
            for status
            in self._app_health_check_state.get_apps_state(app_dependency_names).values())

        if all_instances_completed_with_success and all_apps_completed_private_network_health_check:
            return AppConfigurationStartStatus.START

        return AppConfigurationStartStatus.WAIT

    @staticmethod
    def _build_app_dependencies(app_requests: List[AppRequestInfo]) -> Dict[str, List[str]]:
        _app_dependencies = {app_request.name: list(app_request.app_dependencies) for app_request in app_requests}
        return _app_dependencies



import logging
import os
import platform
import socket


class LoggingFilter(logging.Filter):
    def __init__(self, config_json):
        logging.Filter.__init__(self)
        self.config_json = config_json

    def filter(self, record: logging.LogRecord):
        record.__dict__['colony.application'] = 'sidecar'
        record.__dict__['colony.branch'] = os.getenv('branch', "none")
        record.__dict__['colony.environment'] = self.config_json['environment']
        record.__dict__['colony.hostname'] = self._get_hostname()
        record.__dict__['colony.level'] = record.levelname
        record.__dict__['colony.sandbox_id'] = self.config_json['sandbox_id']
        record.__dict__['colony.threadId'] = record.thread
        record.logger = '{0} {1}'.format(record.filename, record.funcName)
        return record

    @staticmethod
    def _get_hostname():
        n1 = platform.node()
        n2 = socket.gethostname()
        n3 = os.getenv("COMPUTERNAME")
        if n1 == n2 == n3:
            return n1
        elif n1 == n2:
            return n1
        elif n1 == n3:
            return n1
        elif n2 == n3:
            return n2
        else:
            return "local-pc"
